/* $Id: srw-gateway.c,v 1.15 2002/10/16 19:10:23 adam Exp $
   Copyright (C) 2002
   Index Data Aps

This file is part of the SRW tools.

See the file COPYING.
*/

/*
 * TODO:
 *
 * TTL for targets. Separate thread for cleanup.
 * External CCL config.
 * External target config and aliases.
 * XCQL parsing.
 */

/* note that soapH.h defines _REENTRANT so we check for it here */
#ifdef _REENTRANT
#include <pthread.h>
#define USE_THREADS 1
#else
#define USE_THREADS 0
#endif

#include "soapH.h"

#include <yaz/xmalloc.h>
#include <yaz/zoom.h>
#include <yaz/log.h>
#include <yaz/ccl.h>
#include <yaz/yaz-ccl.h>
#include <yaz/options.h>

struct target {
    ZOOM_connection m_c;
    ZOOM_resultset m_r;
    char *m_name;
    char *m_query;
    char *m_schema;
    int m_in_use;
    struct target *next;
};

static struct target *target_list = 0;
#if USE_THREADS
static pthread_mutex_t target_mutex = PTHREAD_MUTEX_INITIALIZER;
#define mylock(x) pthread_mutex_lock(x)
#define myunlock(x) pthread_mutex_unlock(x)
#else
#define mylock(x)
#define myunlock(x)
#endif

struct target *target_use (const char *action, const char *query,
                           const char *schema)
{
    char name[80];
    struct target *l = 0;
    
    if (strlen(action) >= 80)
        action = "localhost:210";
    if (strchr(action, '/') || strchr(action, ':'))
        strcpy (name, action);
    else
    {
        strcpy (name, "localhost/");
        if (*action == '\0')
            strcat (name, "Default");
        else
            strcat (name, action);
    }

    mylock(&target_mutex);

    /* See if we have the target and the same query */
    l = 0;
    if (query)
        for (l = target_list; l; l = l->next)
            if (!l->m_in_use &&
                !strcmp (l->m_name, name) && strcmp(l->m_query, query) &&
                !strcmp (l->m_schema, schema))
                break;

    /* OK, see if we have the target, then.. */
    if (!l)
        for (l = target_list; l; l = l->next)
            if (!strcmp (l->m_name, name) && !l->m_in_use)
                break;

    if (!l)
    {
        /* allright. Have to make a new one */
        l = xmalloc (sizeof(*l));
        l->m_name = xstrdup (name);
        l->m_in_use = 1;
        l->m_c = 0;
        l->m_r = 0;
        l->m_query = xstrdup("");
        l->m_schema = xstrdup("");
        l->next = target_list;
        target_list = l;
    }
    myunlock(&target_mutex);
    return l;
}

static void target_destroy (struct target *t)
{
    struct target **tp;

    mylock(&target_mutex);

    for (tp = &target_list; *tp; tp = &(*tp)->next)
        if (*tp == t)
        {
            *tp = t->next;

            ZOOM_connection_destroy (t->m_c);
            ZOOM_resultset_destroy (t->m_r);

            xfree (t->m_query);
            xfree (t->m_schema);
            xfree (t->m_name);
            xfree (t);
            break;
        }
    myunlock(&target_mutex);
}
    
static void target_leave (struct target *l)
{
    mylock(&target_mutex);
    l->m_in_use = 0;
    myunlock(&target_mutex);
}

static void standalone(struct soap *soap, const char *host, int port,
                       int max_thr, int optimize_level)
{
    struct soap **soap_thr = malloc (sizeof(*soap_thr) * max_thr);
#if USE_THREADS
    pthread_t *tid = malloc (sizeof(pthread_t) * max_thr);
#endif
    int m, s, i;
    int cno = 0;
    int stop = 0;
    char fname[40];
    
    m = soap_bind(soap, 0, port, 100);
    if (m < 0)
    {
        yaz_log (LOG_WARN|LOG_ERRNO, "Cannot bind to %d", port);
        stop = 1;
    }

    for (i = 0; i<max_thr; i++)
        soap_thr[i] = 0;

    while (!stop)
    {
        for (i = 0; i<max_thr; i++)
        {
            s = soap_accept(soap);
            if (s < 0)
                break;
            cno++;
            if (!soap_thr[i])
            {
                soap_thr[i] = soap_new();
                if (!soap_thr[i])
                {
                    stop = 1;
                    break;
                }
            }
            else
            {
#if USE_THREADS
                if (max_thr > 1)          /* static mode for max_thr <= 1 */
                    pthread_join(tid[i], 0);
#endif
                soap_end(soap_thr[i]);
            }

#ifdef DEBUG
            sprintf (fname, "srw.recv.%05d.log", cno);
	    remove (fname);
            soap_set_recv_logfile(soap_thr[i], fname);
            
            sprintf (fname, "srw.sent.%05d.log", cno);
	    remove (fname);
            soap_set_sent_logfile(soap_thr[i], fname);
            
            sprintf (fname, "srw.test.%05d.log", cno);
	    remove (fname);
            soap_set_test_logfile(soap_thr[i], fname);

            yaz_log (LOG_LOG, "starting session %d %ld.%ld.%ld.%ld", cno,
                     (long) (soap->ip>>24) & 0xff,
                     (long) (soap->ip>>16) & 0xff,
                     (long) (soap->ip>>8) & 0xff,
                     (long) soap->ip & 0xff);
#endif
            soap_thr[i]->encodingStyle = 0;
            soap_thr[i]->disable_href = 1;
            soap_thr[i]->socket = s;
            soap_thr[i]->user = &optimize_level;
#if USE_THREADS
            if (max_thr <= 1)
                soap_serve(soap_thr[i]);  /* static mode .. */
            else
                pthread_create(&tid[i], 0, (void*(*)(void*))soap_serve,
                               soap_thr[i]);
#else
            soap_serve(soap_thr[i]);  /* static mode .. */
#endif
        }
    }
#if USE_THREADS
    free (tid);
#endif
    free (soap_thr);
}

static void reconnect (struct target *t)
{
    ZOOM_connection_destroy (t->m_c);
    ZOOM_resultset_destroy (t->m_r);
    
    t->m_r = 0;
    t->m_c = ZOOM_connection_create (0);
    ZOOM_connection_connect (t->m_c, t->m_name, 0);
}

int zs__searchRetrieve(struct soap * soap,
                       zt__cqlType *query,
                       zt__sortSpecType *sort,
                       xsd__integer *startRecord,
                       xsd__integer *maximumRecords,
                       zt__schemaRefType *recordSchema,
                       struct zs__srwSearchRetrieveResponse *res)
{
    const char *msg = 0, *addinfo = 0;
    const char *schema = recordSchema ? *recordSchema : "";
    struct target *t;
    int error = 0;
    int optimize_level = *(int*) (soap->user);
    struct zt__srwSearchRetrieveResponseType 
        *resp = &res->searchRetrieveResponse;
    WRBUF wrbuf_pqf = 0;
    char zurl[81];

    *zurl = '\0';
    yaz_log (LOG_LOG, "HTTP: %s", soap->endpoint);
    if (*soap->endpoint)
    {
        const char *cp = strstr(soap->endpoint, "//");
	if (cp)
            cp = cp+2;             /* skip method// */
	else
            cp = soap->endpoint;
	cp = strstr(cp, "/"); 
	if (cp)
        {
            size_t len;
            cp++;
            len = strlen(cp);
            if (len > 80)
                len = 80;
            if (len)
                memcpy (zurl, cp, len);
            zurl[len] = '\0';
        }
    }
    else
    {
        const char *cp = getenv("PATH_INFO");
        if (cp && cp[0] && cp[1])
        {
            size_t len;
            cp++;  /* skip / */
            len = strlen(cp);
            if (len > 80)
                len = 80;
            if (len)
                memcpy (zurl, cp, len);
            zurl[len] = '\0';
        }
    }
    if (query)
    {
        struct ccl_rpn_node *rpn;
        CCL_parser ccl_parser = ccl_parser_create ();
        struct ccl_token *token_list;
        CCL_bibset bibset = ccl_parser->bibset = ccl_qual_mk();

        ccl_qual_fitem(bibset, "u=1016 t=l,r s=pw", "term");

        ccl_qual_fitem(bibset, "u=4 s=pw", "titleString");
        ccl_qual_fitem(bibset, "u=4 s=2 s=al", "titleWord");

        ccl_qual_fitem(bibset, "u=1003 s=pw", "creatorString");
        ccl_qual_fitem(bibset, "u=1003 s=2 s=al", "creatorWord");
        ccl_qual_fitem(bibset, "u=21 s=pw", "subjectString");
        ccl_qual_fitem(bibset, "u=21 s=2 s=al", "subjectWord");
        ccl_qual_fitem(bibset, "u=1018 s=pw", "publisherString");
        ccl_qual_fitem(bibset, "u=1018 s=2 s=al", "publisherWord");
        ccl_qual_fitem(bibset, "u=30 s=2", "dc.date");

        token_list = ccl_parser_tokenize(ccl_parser, *query);

        rpn = ccl_parser_find (ccl_parser, token_list);
        ccl_token_del (token_list);

        if (ccl_parser->error_code)
        {
            yaz_log (LOG_LOG, "CCL parse %d for query '%s'",
                     ccl_parser->error_code, *query);
            addinfo = ccl_err_msg(ccl_parser->error_code);
            error = 108; /* malformed query */
        }
        else
        {
            wrbuf_pqf = wrbuf_alloc();
            ccl_pquery (wrbuf_pqf, rpn);
            wrbuf_puts (wrbuf_pqf, "");
        }
        ccl_qual_rm(&ccl_parser->bibset);
        ccl_parser_destroy (ccl_parser);
    }
    t = target_use (zurl, wrbuf_pqf ? wrbuf_buf(wrbuf_pqf) : 0, schema);

    if (!t->m_c)
    {
        reconnect(t);
        if (ZOOM_connection_error (t->m_c, &msg, &addinfo))
        {
            yaz_log (LOG_LOG, "%s: connect failed", t->m_name);
            error = 1016;
        }
        else
            yaz_log (LOG_LOG, "%s: connect ok", t->m_name);
    }
    if (t->m_c && !error && wrbuf_pqf)
    {
        if (optimize_level <=1 ||
            strcmp (wrbuf_buf(wrbuf_pqf), t->m_query) ||
            strcmp (schema, t->m_schema))
        {
            /* not the same query: remove result set .. */
            ZOOM_resultset_destroy (t->m_r);
            t->m_r = 0;
        }
        else
        {
            /* same query: retrieve (instead of doing search) */
            if (maximumRecords && *maximumRecords > 0)
            {
                int start = startRecord ? *startRecord : 1;
                yaz_log (LOG_LOG, "%s: present start=%d count=%d",
                         t->m_name, start, *maximumRecords);
                ZOOM_resultset_records (t->m_r, 0, start-1, *maximumRecords);
                error = ZOOM_connection_error (t->m_c, &msg, &addinfo);
                if (error == ZOOM_ERROR_CONNECTION_LOST ||
                    error == ZOOM_ERROR_CONNECT)
                {
                    reconnect (t);
                    if (ZOOM_connection_error (t->m_c, &msg, &addinfo))
                    {
                        yaz_log (LOG_LOG, "%s: connect failed", t->m_name);
                        error = 1016;
                    }
                }
                else if (error)
                {
                    yaz_log (LOG_LOG, "%s: present failed", t->m_name);
                    error = 1016;
                }
            }
        }
        if (!t->m_r) 
        {   /* no result set usable. We must search ... */
            int pass;
            for (pass = 0; pass < 2; pass++)
            {
                char val[30];
                int start = startRecord ? *startRecord : 1;
                int count = maximumRecords ? *maximumRecords : 0;
                
                sprintf (val, "%d", start-1);
                ZOOM_connection_option_set (t->m_c, "start", val);
                
                sprintf (val, "%d", count);
                ZOOM_connection_option_set (t->m_c, "count", val);
                
                ZOOM_connection_option_set (t->m_c, "preferredRecordSyntax", 
                                            "usmarc");
                
                xfree (t->m_query);
                t->m_query = xstrdup (wrbuf_buf(wrbuf_pqf));

                xfree (t->m_schema);
                t->m_schema = xstrdup (schema);
                
                yaz_log (LOG_LOG, "%s: search start=%d count=%d query=%s",
                         t->m_name, start, count, wrbuf_buf(wrbuf_pqf));
                
                t->m_r = ZOOM_connection_search_pqf (t->m_c, t->m_query);
                
                error = ZOOM_connection_error (t->m_c, &msg, &addinfo);
                if (error != ZOOM_ERROR_CONNECTION_LOST &&
                    error != ZOOM_ERROR_CONNECT)
                    break;

                yaz_log (LOG_LOG, "%s: reconnect (search again)", t->m_name);

                /* try once more */
                reconnect(t);

                error = ZOOM_connection_error (t->m_c, &msg, &addinfo);

                if (error)
                {
                    error = 1016;
                    break;
                }
            }
        }
    }
    
    if (error)
    {
        resp->diagnostics.__sizeDiagnostics = 1;
        resp->diagnostics.diagnostic =
            soap_malloc (soap, sizeof(*resp->diagnostics.diagnostic));
        resp->diagnostics.diagnostic[0] =
            soap_malloc (soap, sizeof(**resp->diagnostics.diagnostic));

        resp->diagnostics.diagnostic[0]->code = error;
        if (addinfo)
        {
            resp->diagnostics.diagnostic[0]->details =
                soap_malloc (soap, strlen(addinfo) + 1);
            strcpy (resp->diagnostics.diagnostic[0]->details, addinfo);
        }
    }
    else
    {
        if (t->m_c && t->m_r)
        {
            resp->numberOfRecords = ZOOM_resultset_size(t->m_r);
            
            if (maximumRecords)
            {
                int i;
                int offset = startRecord ? *startRecord -1 : 0;
                resp->records.record =
                    soap_malloc(soap, sizeof(*resp->records.record) *
                                *maximumRecords);
                
                for (i = 0; i < *maximumRecords; i++)
                {
                    int len;
                    const char *xml_rec =
                        ZOOM_record_get(
                            ZOOM_resultset_record (t->m_r, i + offset),
                            "xml", &len);
                    if (!xml_rec)
                    {
                        break;
                    }
                    resp->records.record[i] =
                        soap_malloc (soap, sizeof(**resp->records.record));

                    resp->records.record[i]->recordData =
                        soap_malloc (soap, len+1);
                    memcpy (resp->records.record[i]->recordData,
                            xml_rec, len);
                    resp->records.record[i]->recordData[len] = 0;
                    resp->records.record[i]->recordSchema = "MarcXML";
                }
                resp->records.__sizeRecords = i;
            }
        }
        else
            resp->numberOfRecords = 0;
    }
    resp->resultSetId = "myid";
    if (wrbuf_pqf)
        wrbuf_free (wrbuf_pqf, 1);
    if (optimize_level > 0)
        target_leave(t);
    else
        target_destroy(t);
    return SOAP_OK;
}

int main(int argc, char **argv)
{
    struct soap soap;
    int ret;
    int port = 0;
    int no_threads = 40;
    int optimize_level = 2;
    char *arg;
    const char *host = 0;
            
    while ((ret = options("dO:T:l:hv", argv, argc, &arg)) != -2)
    {
        switch(ret)
        {
        case 0:
            port = atoi(arg);
            break;
        case 'O':
            optimize_level = atoi(arg);
            break;
        case 'T':
            no_threads = atoi(arg);
            if (no_threads < 1 || no_threads > 200)
                no_threads = 40;
            break;
        case 'l':
            yaz_log_init_file(arg);
            break;
        case 'v':
            puts ("Version: $Id: srw-gateway.c,v 1.15 2002/10/16 19:10:23 adam Exp $"
#ifdef DEBUG
            " DEBUG"
#endif
                );
            exit (0);
        case 'h':
            printf ("srw-gateway [-d] [-O<level>] [-T<num>] [-l<file>] "
                    "[<port>]\n");
            printf ("  port  Standalone service port; CGI if port is omitted.\n");
            printf ("  -O    optimize level. >= 1 cache connections, >=2 cache result sets.\n");
#if USE_THREADS
            printf ("  -T    number of threads.\n");
#else
            printf ("  -T    unsupported in this version.\n");
#endif
            printf ("  -l    log to file (instead of stderr).\n");
            printf ("  -v    show version.\n");
            exit (0);
        default:
            fprintf (stderr, "srw-gateway: bad option -%s ; use -h for help\n",
                     arg);
            exit (1);
            break;
            
        }
    }
    soap.encodingStyle = 0;
    soap.disable_href = 1;
    if (port == 0)
    {
        optimize_level = 0;

        yaz_log_init_file("srw.log");
        yaz_log (LOG_LOG, "CGI begin");
        soap_init(&soap);
        soap.user = &optimize_level;

        soap_serve(&soap);

        soap_end(&soap);
        yaz_log (LOG_LOG, "CGI end");
    }
    else
    {
        yaz_log (LOG_LOG, "standalone service on port %d", port);
        
        soap_init(&soap);
        
        standalone(&soap, host, port, no_threads, optimize_level);
    }
    return 0;
}
